from abc import ABC, abstractmethod
from functools import wraps
from typing import TypeAlias, Type, Any, Generic, get_args, TypeVar, Callable, Awaitable

import ommi
import ommi.drivers.add_actions as add_action
import ommi.drivers.find_actions as find_action
import ommi.drivers.schema_actions as schema_action

from ommi.drivers.database_results import AsyncResultWrapper
from ommi.drivers.driver_configs import DriverConfig
from ommi.drivers.driver_types import TConn, TModel
from ommi.models.collections import ModelCollection
from ommi.query_ast import ASTGroupNode

DriverName: TypeAlias = str
DriverNiceName: TypeAlias = str
Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool
ConnectionProtocol = TypeVar("ConnectionProtocol")


class AbstractDatabaseDriver(Generic[TConn, TModel], ABC):
    __drivers__: "dict[DriverName, Type[AbstractDatabaseDriver]]"
    driver_name: DriverName

    # ---------------------------- #
    # Connection Management        #
    # ---------------------------- #

    @property
    @abstractmethod
    def connection(self) -> Any:
        ...

    @property
    @abstractmethod
    def connected(self) -> bool:
        ...

    @abstractmethod
    def disconnect(self) -> AsyncResultWrapper[bool]:
        ...

    # ---------------------------- #
    # Actions                      #
    # ---------------------------- #

    @property
    @abstractmethod
    def add(self, *items: TModel) -> "add_action.AddAction[TConn, TModel]":
        ...

    @abstractmethod
    def find(self, *predicates: Predicate) -> "find_action.FindAction[TConn, TModel]":
        ...

    @abstractmethod
    def schema(
        self, model_collection: ModelCollection[Type[TModel]] | None = None
    ) -> "schema_action.SchemaAction[TConn, TModel]":
        ...

    # ---------------------------- #
    # Driver Collection Management #
    # ---------------------------- #

    @classmethod
    @abstractmethod
    def add_driver(cls, driver: "Type[AbstractDatabaseDriver]"):
        ...

    @classmethod
    @abstractmethod
    def disable_driver(cls, name: DriverName):
        ...

    # ---------------------------- #
    # Configuration                #
    # ---------------------------- #

    @classmethod
    @abstractmethod
    async def from_config(cls, config: DriverConfig) -> "AbstractDatabaseDriver":
        ...


class DatabaseDriver(AbstractDatabaseDriver[TConn, TModel], ABC):
    __drivers__ = {}

    def __init_subclass__(cls, **kwargs):
        cls.driver_name = kwargs.pop(
            "driver_name", getattr(cls, "driver_name", cls.__name__)
        )
        cls.nice_name = kwargs.pop("nice_name", getattr(cls, "nice_name", cls.__name__))

        super().__init_subclass__(**kwargs)
        cls.add_driver(cls)

    def __init__(self, connection: TConn):
        self._connection = connection
        self._connected = True

    def __enter__(self):
        if hasattr(self, "_driver_context"):
            return self

        self._driver_context = ommi.driver_context.UseDriver(self)
        return self._driver_context.__enter__()

    def __exit__(self, *args):
        if not hasattr(self, "_driver_context"):
            return

        context = self._driver_context
        del self._driver_context
        try:
            return context.__exit__(*args)
        except ValueError:
            return True

    async def __aenter__(self):
        self.__enter__()
        return self

    async def __aexit__(self, *args):
        await self.disconnect().raise_on_errors()
        self.__exit__(*args)
        return

    @property
    def connection(self) -> TConn:
        return self._connection

    @property
    def connected(self) -> bool:
        return self._connected

    @classmethod
    def add_driver(cls, driver: "Type[AbstractDatabaseDriver]"):
        cls.__drivers__[driver.driver_name] = driver

    @classmethod
    def disable_driver(cls, name: DriverName):
        cls.__drivers__.pop(name, None)


class ConnectionFromConfigContextManager:
    def __init__(self, awaitable: Awaitable[AbstractDatabaseDriver]):
        self.awaitable = awaitable
        self.connection = None

    def __await__(self):
        return self.awaitable.__await__()

    async def __aenter__(self):
        self.connection = await self.awaitable
        await self.connection.__aenter__()
        return self.connection

    async def __aexit__(self, *args):
        return await self.connection.__aexit__(*args)


def connection_context_manager(
    func: Callable[[DriverConfig], Awaitable[AbstractDatabaseDriver]]
):
    @wraps(func)
    def wrapper(cls, config: DriverConfig) -> ConnectionFromConfigContextManager:
        return ConnectionFromConfigContextManager(func(cls, config))

    return wrapper


def enforce_connection_protocol(driver: Type[AbstractDatabaseDriver]):
    init = driver.__init__
    connection_protocol = get_args(driver.__orig_bases__[0])[0]
    if connection_protocol is ConnectionProtocol:
        raise TypeError(
            f"Expected a connection protocol to be defined on the driver class {driver.__qualname__}."
        )

    @wraps(driver.__init__)
    def __init__(self, connection, *args, **kwargs):
        if connection_protocol and not isinstance(connection, connection_protocol):
            raise TypeError(
                f"Expected connection implementing the {connection_protocol.__qualname__} protocol, {type(connection)} does not."
            )

        init(self, connection, *args, **kwargs)

    driver.__init__ = __init__
    return driver



from typing import Sequence, Type, Any
from tramp.optionals import Optional

from ommi.drivers.database_results import async_result
from ommi.drivers.delete_actions import DeleteAction
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.ext.drivers.mongodb.utils import build_pipeline, process_ast, Query
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode

Predicate = ASTGroupNode | type(OmmiModel) | bool


class MongoDBDeleteAction(DeleteAction[MongoDBConnection, OmmiModel]):
    def __init__(
        self, connection: MongoDBConnection, predicates: Sequence[Predicate], database
    ):
        super().__init__(connection, predicates)
        self._db = database
        self._is_replica_set_result: Optional[bool] = Optional.Nothing

    @property
    async def _is_replica_set(self) -> bool:
        match self._is_replica_set_result:
            case Optional.Nothing:
                result = await self._connection.local.system.replset.find().to_list(1)
                self._is_replica_set_result = Optional.Some(len(result) != 0)

            case Optional.Some(result):
                return result

    @async_result
    async def delete(self) -> bool:
        query = process_ast(when(*self._predicates))
        if query.collections:
            return await self._join_delete(query)

        return await self._delete(query.collection, query.match)

    async def _delete(
        self, model: Type[OmmiModel], match: list[dict[str, Any]]
    ) -> bool:
        await self._db[model.__ommi__.model_name].delete_many(
            {"$and": match} if match else {},
        )
        return True

    async def _join_delete(self, query: Query) -> bool:
        async with await self._connection.start_session() as session:
            if await self._is_replica_set:
                return await self._transaction_delete(query, session)

            return await self._do_join_delete(query, session)

    async def _do_join_delete(self, query: Query, session=None) -> bool:
        pipeline, model = build_pipeline(query)
        documents_to_delete = self._db[model.__ommi__.model_name].aggregate(
            pipeline, session=session
        )
        await self._db[model.__ommi__.model_name].delete_many(
            {"_id": {"$in": [doc["_id"] async for doc in documents_to_delete]}},
            session=session,
        )
        return True

    async def _transaction_delete(self, query: Query, session) -> bool:
        async with session.start_transaction():
            return await self._do_join_delete(query, session)



from dataclasses import dataclass, field as dc_field
from typing import Type, Any, TypeAlias, TypedDict

from ommi.models import OmmiModel
from ommi.query_ast import (
    ASTGroupNode,
    ResultOrdering,
    ASTGroupFlagNode,
    ASTComparisonNode,
    ASTOperatorNode,
    ASTLogicalOperatorNode,
    ASTLiteralNode,
    ASTReferenceNode,
)

logical_operator_mapping = {
    ASTLogicalOperatorNode.AND: "$and",
    ASTLogicalOperatorNode.OR: "$or",
}

operator_mapping = {
    ASTOperatorNode.EQUALS: "$eq",
    ASTOperatorNode.NOT_EQUALS: "$ne",
    ASTOperatorNode.GREATER_THAN: "$gt",
    ASTOperatorNode.GREATER_THAN_OR_EQUAL: "$gte",
    ASTOperatorNode.LESS_THAN: "$lt",
    ASTOperatorNode.LESS_THAN_OR_EQUAL: "$lte",
}

# Field references need to be on the left, so if they're on the right the operator needs to be flipped
flipped_operator_mapping = operator_mapping | {
    ASTOperatorNode.GREATER_THAN: "$lt",
    ASTOperatorNode.GREATER_THAN_OR_EQUAL: "$lte",
    ASTOperatorNode.LESS_THAN: "$gt",
    ASTOperatorNode.LESS_THAN_OR_EQUAL: "$gte",
}

LocalField: TypeAlias = str
ForeignField: TypeAlias = str

LookupStage = TypedDict(
    "LookupStage",
    {
        "$lookup": TypedDict(
            "LookupStageFields",
            {
                "from": str,
                "localField": str,
                "foreignField": str,
                "as": str,
            },
        )
    },
)


ProjectStage = TypedDict(
    "ProjectStage",
    {
        "$project": dict[str, int],
    },
)


UnwindStage = TypedDict(
    "UnwindStage",
    {
        "$unwind": TypedDict(
            "UnwindStageFields",
            {
                "path": str,
                "preserveNullAndEmptyArrays": bool,
            },
        )
    },
)


LookupStages: TypeAlias = list[LookupStage]
UnwindStages: TypeAlias = list[UnwindStage]


@dataclass
class Query:
    collection: Type[OmmiModel] | None = None
    collections: list[Type[OmmiModel]] = dc_field(default_factory=list)
    match: list[Any] = dc_field(default_factory=list)

    sorts: list[ASTReferenceNode] = dc_field(default_factory=list)
    max_results: int = 0
    results_page: int = 0

    def add_collection(self, *collections: Type[OmmiModel]) -> None:
        if not self.collection:
            self.collection, *collections = collections

        if collections:
            self.collections.extend(
                collection
                for collection in collections
                if collection not in self.collections and collection != self.collection
            )


def model_to_dict(model: OmmiModel, *, preserve_pk: bool = False) -> dict[str, Any]:
    fields = list(model.__ommi__.fields.values())
    pks = model.get_primary_key_fields()
    return {
        field.get("store_as"): getattr(model, field.get("field_name"))
        for field in fields
        if field not in pks
        or preserve_pk
        or getattr(model, field.get("field_name")) is not None
    }


def process_ast(ast: ASTGroupNode) -> Query:
    query = Query()
    group_stack = [query.match]
    logical_operator_stack = [ASTLogicalOperatorNode.AND]
    node_stack = [iter(ast)]
    while node_stack:
        match next(node_stack[~0], None):
            case ASTReferenceNode(field=None, model=model):
                query.add_collection(model)

            case ASTGroupNode() as group:
                node_stack.append(iter(group))

            case ASTComparisonNode(left, right, op):
                expression, collections = _process_comparison_ast(
                    left, op, right, query.collection
                )
                group_stack[~0].append(expression)
                query.add_collection(*collections)

            case ASTLogicalOperatorNode() as op:
                if len(node_stack) <= 1:
                    continue

                if logical_operator_stack[~0] is None:
                    logical_operator_stack[~0] = op
                    group_stack[~1].append(
                        {logical_operator_mapping[op]: group_stack[~0]}
                    )

                elif logical_operator_stack[~0] != op:
                    logical_operator_stack.append(op)
                    group_stack.append([])
                    group_stack[~1].append(
                        {logical_operator_mapping[op]: group_stack[~0]}
                    )

            case ASTGroupFlagNode.OPEN:
                if len(node_stack) > 1:
                    group_stack.append([])
                    logical_operator_stack.append(None)

            case ASTGroupFlagNode.CLOSE:
                group_stack.pop()

            case None:
                node_stack.pop()

            case node:
                raise TypeError(f"Unexpected node type: {node}")

    if ast.sorting:
        query.sorts = ast.sorting

    if ast.max_results:
        query.max_results = ast.max_results

    if ast.results_page:
        query.results_page = ast.results_page

    return query


def build_pipeline(query: Query) -> tuple[list[dict[str, Any]], Type[OmmiModel]]:
    pipeline = []
    if query.match:
        pipeline.append({"$match": {"$and": query.match}})

    if query.sorts:
        pipeline.append(_create_sort_stage(query.sorts))

    if query.max_results > 0:
        pipeline.append(_create_limit_stage(query.max_results))

        if query.results_page:
            pipeline.append(_create_skip_stage(query.max_results, query.results_page))

    if len(query.collections):
        lookups, unwind, project = create_lookup_stages(
            query.collection, query.collections
        )
        pipeline = [*lookups, *unwind, *pipeline, project]

    return pipeline, query.collection


def _create_sort_stage(sorts: list[ASTReferenceNode]) -> dict[str, Any]:
    return {
        "$sort": {
            ref.field.name: 1 if ref.ordering == ResultOrdering.ASCENDING else -1
            for ref in sorts
        }
    }


def create_lookup_stages(
    model: Type[OmmiModel], collections: list[Type[OmmiModel]]
) -> tuple[LookupStages, UnwindStages, ProjectStage]:
    lookups = []
    project = {
        "$project": (hide := {}),
    }
    unwind = []
    for collection in collections:
        hide[f"__join__{collection.__ommi__.model_name}"] = 0

        refs = _get_reference_fields(model, collection)
        model_name = model.__ommi__.model_name.lower()
        lookups.append(
            {
                "$lookup": {
                    "from": collection.__ommi__.model_name,
                    "let": {
                        f"{model_name}_{local_field}": f"${local_field}"
                        for local_field, _ in refs
                    },
                    "pipeline": [
                        {
                            "$match": {
                                "$expr": {
                                    "$and": [
                                        {
                                            "$eq": [
                                                f"${foreign_field}",
                                                f"$${model_name}_{local_field}",
                                            ]
                                        }
                                        for local_field, foreign_field in refs
                                    ]
                                }
                            }
                        }
                    ],
                    "as": f"__join__{collection.__ommi__.model_name}",
                }
            }
        )

        unwind.append(
            {
                "$unwind": {
                    "path": f"$__join__{collection.__ommi__.model_name}",
                    "preserveNullAndEmptyArrays": True,
                }
            }
        )

    return lookups, unwind, project


def _create_limit_stage(max_results: int) -> dict[str, Any]:
    return {"$limit": max_results}


def _create_skip_stage(max_results: int, results_page: int) -> dict[str, Any]:
    return {"$skip": max_results * results_page}


def _get_reference_fields(
    model: Type[OmmiModel], collection: Type[OmmiModel]
) -> tuple[tuple[LocalField, ForeignField], ...]:
    if ref := collection.__ommi__.references.get(model):
        return tuple(
            (r.from_field.get("store_as"), r.to_field.get("store_as")) for r in ref
        )

    ref = model.__ommi__.references[collection]
    return tuple(
        (r.from_field.get("store_as"), r.to_field.get("store_as")) for r in ref
    )


def _process_comparison_ast(
    left: ASTLiteralNode | ASTReferenceNode,
    op: ASTOperatorNode,
    right: ASTLiteralNode | ASTReferenceNode,
    querying_model: Type[OmmiModel] | None,
) -> tuple[dict[str, Any], list[Type[OmmiModel]]]:
    collections = []
    match left, right:
        case (
            ASTReferenceNode(field=field, model=model),
            ASTLiteralNode(value=value),
        ):
            collections.append(model)

            name = field.metadata.get("store_as")
            if (querying_model and model != querying_model) or (
                not querying_model and collections and model != collections[0]
            ):
                name = f"__join__{model.__ommi__.model_name}.{name}"

            expr = {
                name: (
                    value
                    if op == ASTOperatorNode.EQUALS
                    else {operator_mapping[op]: value}
                )
            }

        case (
            ASTLiteralNode(value=value),
            ASTReferenceNode(field=field, model=model),
        ):
            collections.append(model)

            name = field.metadata.get("store_as")
            if model != querying_model or (
                not querying_model and model != collections[0]
            ):
                name = f"__join__{model.__ommi__.model_name}.{name}"

            expr = {
                name: (
                    value
                    if op == ASTOperatorNode.EQUALS
                    else {flipped_operator_mapping[op]: value}
                )
            }

        case _:
            raise TypeError(f"Unexpected node type: {left} or {right}")

    return expr, collections



from ommi.drivers import DatabaseDriver
from ommi.driver_context import active_driver, use_driver
from ommi.models import ommi_model
from ommi.models.field_metadata import Auto, Key, FieldType, StoreAs



from typing import Any, Sequence

from ommi.drivers.database_results import async_result
from ommi.drivers.set_fields_actions import SetFieldsAction
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.ext.drivers.mongodb.utils import process_ast, create_lookup_stages
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode

Predicate = ASTGroupNode | type[OmmiModel] | bool


class MongoDBSetFieldsAction(SetFieldsAction[MongoDBConnection, OmmiModel]):
    def __init__(
        self, connection: MongoDBConnection, predicates: Sequence[Predicate], database
    ):
        super().__init__(connection, predicates)
        self._db = database

    @async_result
    async def set_fields(self, **kwargs: Any) -> bool:
        query = process_ast(when(*self._predicates))
        pipeline = [
            {
                "$match": {
                    "$and": query.match,
                },
            },
            {
                "$set": {
                    query.collection.__ommi__.fields[name].get("store_as"): value
                    for name, value in kwargs.items()
                },
            },
        ]

        if query.collections:
            lookups, unwind, project = create_lookup_stages(
                query.collection, query.collections
            )
            pipeline = [*lookups, *unwind, *pipeline, project]

        pipeline.append(
            {
                "$merge": {
                    "into": query.collection.__ommi__.model_name,
                    "on": "_id",
                    "whenMatched": "replace",
                },
            }
        )

        await self._db[query.collection.__ommi__.model_name].aggregate(
            pipeline
        ).to_list(1)
        return True



import sqlite3
from typing import Type, Iterable, Any

from ommi.drivers.database_results import async_result
from ommi.drivers.schema_actions import SchemaAction
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.models.field_metadata import FieldMetadata
from ommi.models import OmmiModel
from ommi.models.models import get_collection
from tramp.optionals import Optional


class SQLiteSchemaAction(SchemaAction[SQLiteConnection, OmmiModel]):
    type_mapping = {
        int: "INTEGER",
        str: "TEXT",
        float: "REAL",
        bool: "INTEGER",
    }

    @async_result
    async def create_models(self) -> Iterable[Type[OmmiModel]]:
        session = self._connection.cursor()
        models = get_collection(
            Optional.Some(self._model_collection)
            if self._model_collection
            else Optional.Nothing
        ).models

        try:
            for model in models:
                self._create_table(model, session)

        except:
            self._connection.rollback()
            raise

        else:
            return models

        finally:
            session.close()

    @async_result
    async def delete_models(self) -> Iterable[Type[OmmiModel]]:
        session = self._connection.cursor()
        models = get_collection(
            Optional.Some(self._model_collection)
            if self._model_collection
            else Optional.Nothing
        ).models

        try:
            for model in models:
                session.execute(f"DROP TABLE IF EXISTS {model.__ommi__.model_name};")

        except:
            self._connection.rollback()
            raise

        else:
            return models

        finally:
            session.close()

    def _create_table(self, model: Type[OmmiModel], session: sqlite3.Cursor):
        pks = model.get_primary_key_fields()
        columns = ", ".join(
            self._build_column(field) for field in model.__ommi__.fields.values()
        )
        session.execute(
            f"CREATE TABLE IF NOT EXISTS {model.__ommi__.model_name}"
            f"("
            f"{columns}, "
            f"PRIMARY KEY ({', '.join(pk.get('store_as') for pk in pks)})"
            f");"
        )

    def _build_column(self, field: FieldMetadata) -> str:
        column = [
            field.get("store_as"),
            self._get_sqlite_type(field.get("field_type")),
        ]

        return " ".join(column)

    def _get_sqlite_type(self, type_: Type[Any]) -> str:
        return self.type_mapping.get(type_, "TEXT")



from typing import Sequence, Type

from ommi.drivers.count_actions import CountAction
from ommi.drivers.delete_actions import DeleteAction
from ommi.drivers.driver_types import TModel, TConn
from ommi.drivers.fetch_actions import FetchAction
from ommi.drivers.find_actions import FindAction
from ommi.drivers.set_fields_actions import SetFieldsAction
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.ext.drivers.mongodb.count_action import MongoDBCountAction
from ommi.ext.drivers.mongodb.delete_action import MongoDBDeleteAction
from ommi.ext.drivers.mongodb.fetch_action import MongoDBFetchAction
from ommi.ext.drivers.mongodb.set_fields_action import MongoDBSetFieldsAction
from ommi.models import OmmiModel
from ommi.query_ast import ASTGroupNode

Predicate = ASTGroupNode | Type[OmmiModel] | bool


class MongoDBFindAction(FindAction[MongoDBConnection, OmmiModel]):
    _count_action = MongoDBCountAction
    _delete_action = MongoDBDeleteAction
    _fetch_action = MongoDBFetchAction
    _set_fields_action = MongoDBSetFieldsAction

    def __init__(
        self, connection: MongoDBConnection, predicates: Sequence[Predicate], database
    ):
        super().__init__(connection, predicates)
        self._db = database

    @property
    def count(self) -> CountAction[TConn, TModel]:
        return self._count_action(self._connection, self._predicates, self._db)

    @property
    def delete(self) -> DeleteAction[TConn, TModel]:
        return self._delete_action(self._connection, self._predicates, self._db)

    @property
    def fetch(self) -> FetchAction[TConn, TModel]:
        return self._fetch_action(self._connection, self._predicates, self._db)

    @property
    def set(self) -> SetFieldsAction[TConn, TModel]:
        return self._set_fields_action(self._connection, self._predicates, self._db)



from typing import Any

import psycopg

from ommi.drivers.database_results import async_result
from ommi.drivers.set_fields_actions import SetFieldsAction
from ommi.ext.drivers.postgresql.connection_protocol import PostgreSQLConnection
from ommi.ext.drivers.postgresql.utils import build_query, create_join_comparison
from ommi.ext.drivers.postgresql.utils import generate_joins
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode


class PostgreSQLSetFieldsAction(SetFieldsAction[PostgreSQLConnection, OmmiModel]):
    @async_result
    async def set_fields(self, **kwargs: Any) -> bool:
        ast = when(*self._predicates)
        session = self._connection.cursor()
        await self._update(ast, kwargs, session)
        return True

    async def _update(
        self,
        ast: ASTGroupNode,
        set_fields: dict[str, Any],
        session: psycopg.AsyncCursor,
    ):
        query = build_query(ast)
        fields = query.model.__ommi__.fields
        where = query.where
        query_builder = [
            f"UPDATE {query.model.__ommi__.model_name}",
            f"SET",
            ", ".join(
                f"{fields[name].get('store_as')} = %s" for name in set_fields.keys()
            ),
        ]
        if query.models:
            from_join = query.models.pop(0)
            from_join_comparison = f"{create_join_comparison(query.model, from_join)}"
            query_builder.append(f"FROM {from_join.__ommi__.model_name}")
            query_builder.extend(generate_joins(query.model, query.models))

            if where:
                where = f"{from_join_comparison} AND ({where})"

            else:
                where = from_join_comparison

        if where:
            query_builder.append(f"WHERE {where}")

        await session.execute(
            f"{' '.join(query_builder)};".encode(),
            (*set_fields.values(), *query.values),
        )



import sqlite3
from typing import Protocol, runtime_checkable


@runtime_checkable
class SQLiteConnection(Protocol):
    def close(self) -> None:
        ...

    def cursor(self) -> sqlite3.Cursor:
        ...

    def rollback(self) -> None:
        ...



from dataclasses import dataclass, field as dc_field

from ommi.models.collections import get_global_collection
from ommi.models.references import LazyReferenceBuilder
import ommi.models.field_metadata


@dataclass
class OmmiMetadata:
    model_name: str
    fields: "dict[str, ommi.models.field_metadata.FieldMetadata]"
    references: LazyReferenceBuilder
    collection: "ommi.models.collections.ModelCollection" = dc_field(
        default_factory=get_global_collection
    )

    def clone(self, **kwargs) -> "OmmiMetadata":
        return OmmiMetadata(
            **{name: kwargs.get(name, value) for name, value in vars(self).items()}
        )



from typing import Generic, Type, Sequence, TypeAlias

from ommi.drivers.count_actions import CountAction
from ommi.drivers.delete_actions import DeleteAction
from ommi.drivers.driver_types import TConn, TModel
from ommi.drivers.fetch_actions import FetchAction
from ommi.drivers.set_fields_actions import SetFieldsAction
from ommi.query_ast import ASTGroupNode

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


class FindAction(Generic[TConn, TModel]):
    _count_action: Type[CountAction[TConn, TModel]]
    _delete_action: Type[DeleteAction[TConn, TModel]]
    _fetch_action: Type[FetchAction[TConn, TModel]]
    _set_fields_action: Type[SetFieldsAction[TConn, TModel]]

    def __init__(
        self,
        connection: TConn,
        predicates: Sequence[Predicate],
    ):
        self._connection = connection
        self._predicates = predicates

    @property
    def count(self) -> CountAction[TConn, TModel]:
        return self._count_action(self._connection, self._predicates)

    @property
    def delete(self) -> DeleteAction[TConn, TModel]:
        return self._delete_action(self._connection, self._predicates)

    @property
    def fetch(self) -> FetchAction[TConn, TModel]:
        return self._fetch_action(self._connection, self._predicates)

    @property
    def set(self) -> SetFieldsAction[TConn, TModel]:
        return self._set_fields_action(self._connection, self._predicates)



import psycopg
from dataclasses import dataclass
from typing import Type, cast, TypeAlias

from ommi.drivers import DatabaseDriver, DriverConfig
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.drivers.drivers import enforce_connection_protocol, connection_context_manager
from ommi.ext.drivers.postgresql.add_action import PostgreSQLAddAction
from ommi.ext.drivers.postgresql.connection_protocol import PostgreSQLConnection
from ommi.ext.drivers.postgresql.find_action import PostgreSQLFindAction
from ommi.ext.drivers.postgresql.schema_action import PostgreSQLSchemaAction
from ommi.models.collections import ModelCollection
from ommi.models import OmmiModel
from ommi.query_ast import ASTGroupNode

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


@dataclass
class PostgreSQLConfig(DriverConfig):
    host: str
    port: int
    database_name: str
    username: str
    password: str

    def to_uri(self) -> str:
        return f"postgresql://{self.username}:{self.password}@{self.host}:{self.port}/{self.database_name}"


@enforce_connection_protocol
class PostgreSQLDriver(
    DatabaseDriver[PostgreSQLConnection, OmmiModel],
    driver_name="postgresql",
    nice_name="PostgreSQL",
):
    @async_result
    async def disconnect(self) -> bool:
        await self._connection.close()
        self._connected = False
        return True

    @property
    def add(self) -> PostgreSQLAddAction:
        return PostgreSQLAddAction(self._connection)

    def find(self, *predicates: Predicate) -> PostgreSQLFindAction:
        return PostgreSQLFindAction(self._connection, predicates)

    def schema(
        self, model_collection: ModelCollection[Type[OmmiModel]] | None = None
    ) -> PostgreSQLSchemaAction:
        return PostgreSQLSchemaAction(self._connection, model_collection)

    @classmethod
    @connection_context_manager
    async def from_config(cls, config: PostgreSQLConfig) -> "PostgreSQLDriver":
        connection = await psycopg.AsyncConnection.connect(config.to_uri())
        return cls(cast(PostgreSQLConnection, connection))



from typing import Generic, Sequence, Any, TypeAlias, Type
from abc import ABC, abstractmethod

from ommi.drivers.database_results import AsyncResultWrapper
from ommi.drivers.driver_types import TConn, TModel
from ommi.query_ast import ASTGroupNode

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


class SetFieldsAction(Generic[TConn, TModel], ABC):
    def __init__(self, connection: TConn, predicates: Sequence[Predicate]):
        self._connection = connection
        self._predicates = predicates

    def __call__(self, **kwargs: Any) -> AsyncResultWrapper[bool]:
        return self.set_fields(**kwargs)

    @abstractmethod
    def set_fields(self, **kwargs: Any) -> AsyncResultWrapper[bool]:
        ...



from ommi.drivers.find_actions import FindAction
from ommi.ext.drivers.postgresql.connection_protocol import PostgreSQLConnection
from ommi.ext.drivers.postgresql.count_action import PostgreSQLCountAction
from ommi.ext.drivers.postgresql.delete_action import PostgreSQLDeleteAction
from ommi.ext.drivers.postgresql.fetch_action import PostgreSQLFetchAction
from ommi.ext.drivers.postgresql.set_fields_action import PostgreSQLSetFieldsAction
from ommi.models import OmmiModel


class PostgreSQLFindAction(FindAction[PostgreSQLConnection, OmmiModel]):
    _count_action = PostgreSQLCountAction
    _delete_action = PostgreSQLDeleteAction
    _fetch_action = PostgreSQLFetchAction
    _set_fields_action = PostgreSQLSetFieldsAction



from typing import Protocol, runtime_checkable
import psycopg


@runtime_checkable
class PostgreSQLConnection(Protocol):
    def cursor(self) -> psycopg.AsyncCursor:
        ...

    async def close(self) -> None:
        ...

    async def commit(self) -> None:
        ...

    async def rollback(self) -> None:
        ...



import sqlite3
from typing import Iterable

from ommi.drivers.add_actions import AddAction
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.models import OmmiModel


class SQLiteAddAction(AddAction[SQLiteConnection, OmmiModel]):
    @async_result
    async def items(self, *items: TModel) -> Iterable[TModel]:
        session = self._connection.cursor()
        try:
            for item in items:
                self._insert(item, session)
                self._sync_with_last_inserted(item, session)

        except Exception:
            self._connection.rollback()
            raise

        else:
            return items

        finally:
            session.close()

    def _insert(self, item: OmmiModel, session: sqlite3.Cursor):
        fields = list(item.__ommi__.fields.values())
        data = {
            field.get("store_as"): getattr(item, field.get("field_name"))
            for field in fields
        }
        qs = ", ".join(["?"] * len(data))
        columns = ", ".join(data.keys())
        values = tuple(data.values())
        session.execute(
            f"INSERT INTO {item.__ommi__.model_name} ({columns}) VALUES ({qs});",
            values,
        )

    def _sync_with_last_inserted(self, item: OmmiModel, session: sqlite3.Cursor):
        pks = item.get_primary_key_fields()
        if len(pks) != 1:
            return

        pk = pks[0].get("store_as")
        result = session.execute("SELECT last_insert_rowid();").fetchone()
        setattr(item, pk, result[0])



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Any, Type

from tramp.results import Result

import ommi
import ommi.drivers.drivers
import ommi.query_ast

T = TypeVar("T")


class LazyQueryField(ABC):
    def __init__(
        self,
        query: "ommi.query_ast.ASTGroupNode",
        driver: "ommi.drivers.drivers.AbstractDatabaseDriver | None" = None,
    ):
        self._query = query
        self._driver = driver

        self._cache = Result.Error(ValueError("Not cached yet"))

    def __await__(self):
        return self.value.__await__()

    def __get_pydantic_core_schema__(self, *_):
        import pydantic_core

        return pydantic_core.core_schema.no_info_plain_validator_function(
            function=self.__pydantic_validator
        )

    @staticmethod
    def __pydantic_validator(value):
        if value is None:
            return value

        if isinstance(value, LazyQueryField):
            return value

        raise TypeError(f"Expected LazyQueryField, got {type(value)}")

    @abstractmethod
    async def get(self, default=None):
        ...

    async def refresh(self) -> None:
        with Result.build() as builder:
            builder.set(await self._fetch())

        self._cache = builder.result

    async def refresh_if_needed(self) -> None:
        match self._cache:
            case Result.Error():
                await self.refresh()

    @property
    @abstractmethod
    async def result(self):
        ...

    @property
    @abstractmethod
    async def value(self):
        ...

    @abstractmethod
    async def _fetch(self):
        ...

    async def _get_result(self):
        match self._cache:
            case Result.Value() as result:
                return result

            case Result.Error():
                self._cache = await self._fetch()
                return self._cache

    def _get_driver(self):
        return self._driver or ommi.active_driver.get()

    @classmethod
    @abstractmethod
    def create(
        cls, model: "ommi.models.OmmiModel", args: tuple[Any, ...]
    ) -> "LazyQueryField":
        ...


class LazyLoadTheRelated(Generic[T], LazyQueryField):
    async def get(self, default: T | None = None) -> T | None:
        return (await self.result).value_or(default)

    @property
    async def result(self) -> Result[T]:
        return await self._get_result()

    @property
    async def value(self) -> T:
        return (await self.result).value

    async def _fetch(self):
        with Result.build() as builder:
            builder.set(await self._get_driver().find(self._query.limit(1)).fetch.one())

        return builder.result

    @classmethod
    def create(
        cls, model: "ommi.models.OmmiModel", args: tuple[Any, ...]
    ) -> "LazyQueryField":
        contains, *_ = args
        return cls(_build_query(model, contains))


class LazyLoadEveryRelated(Generic[T], LazyQueryField):
    async def get(self, default: list[T] | None = None) -> list[T] | None:
        return (await self.result).value_or(default)

    @property
    async def result(self) -> Result[list[T]]:
        return await self._get_result()

    @property
    async def value(self) -> list[T]:
        return (await self.result).value

    async def _fetch(self):
        with Result.build() as builder:
            builder.set(await self._get_driver().find(self._query).fetch.all())

        return builder.result

    @classmethod
    def create(
        cls, model: "ommi.models.OmmiModel", args: tuple[Any, ...]
    ) -> "LazyQueryField":
        contains, *_ = args
        return cls(_build_query(model, contains))


def _build_query(
    model: "ommi.models.OmmiModel", contains: "Type[ommi.models.OmmiModel]"
) -> "ommi.query_ast.ASTGroupNode":
    if refs := model.__ommi__.references.get(contains):
        return ommi.query_ast.when(
            *(
                getattr(r.to_model, r.to_field.get("field_name"))
                == getattr(model, r.from_field.get("field_name"))
                for r in refs
            )
        )

    if refs := contains.__ommi__.references.get(type(model)):
        return ommi.query_ast.when(
            *(
                getattr(r.from_model, r.from_field.get("field_name"))
                == getattr(model, r.to_field.get("field_name"))
                for r in refs
            )
        )

    raise RuntimeError(
        f"No reference found between models {type(model)} and {contains}"
    )



from typing import Type, Generic

from ommi.drivers.driver_types import TModel

_global_collection = None


class ModelCollection(Generic[TModel]):
    def __init__(self):
        self.models = set()

    def add(self, model: "Type[TModel]"):
        self.models.add(model)

    def __repr__(self):
        return f"<{type(self).__name__}: contains {len(self.models)} model{'' if len(self.models) == 1 else 's'}>"


def get_global_collection() -> ModelCollection:
    global _global_collection
    if not _global_collection:
        _global_collection = ModelCollection()

    return _global_collection



from dataclasses import dataclass, field as dc_field
from typing import Type, Any

from ommi.models import OmmiModel
from ommi.query_ast import (
    ASTGroupNode,
    ResultOrdering,
    ASTGroupFlagNode,
    ASTComparisonNode,
    ASTOperatorNode,
    ASTLogicalOperatorNode,
    ASTLiteralNode,
    ASTReferenceNode,
)

logical_operator_mapping = {
    ASTLogicalOperatorNode.AND: "AND",
    ASTLogicalOperatorNode.OR: "OR",
}

operator_mapping = {
    ASTOperatorNode.EQUALS: "=",
    ASTOperatorNode.NOT_EQUALS: "!=",
    ASTOperatorNode.GREATER_THAN: ">",
    ASTOperatorNode.GREATER_THAN_OR_EQUAL: ">=",
    ASTOperatorNode.LESS_THAN: "<",
    ASTOperatorNode.LESS_THAN_OR_EQUAL: "<=",
}


@dataclass
class SelectQuery:
    limit: int = 0
    model: Type[OmmiModel] | None = None
    models: list[Type[OmmiModel]] = dc_field(default_factory=list)
    offset: int = 0
    order_by: dict[str, ResultOrdering] = dc_field(default_factory=dict)
    values: list[Any] = dc_field(default_factory=list)
    where: str = ""

    def add_model(self, *models: Type[OmmiModel]):
        if not self.model:
            self.model, *models = models

        if models:
            self.models.extend(
                m for m in models if m not in self.models and m != self.model
            )


def build_query(ast: ASTGroupNode) -> SelectQuery:
    query = SelectQuery(
        limit=ast.max_results,
        offset=ast.results_page * ast.max_results,
        order_by=_process_ordering(ast.sorting),
    )
    where = []
    node_stack = [iter(ast)]
    while node_stack:
        match next(node_stack[~0], None):
            case ASTGroupNode() as group:
                node_stack.append(iter(group))

            case ASTReferenceNode(None, model):
                query.add_model(model)

            case ASTReferenceNode(field, model):
                name = f"{model.__ommi__.model_name}.{field.metadata.get('store_as')}"
                where.append(name)
                query.add_model(model)

            case ASTLiteralNode(value):
                where.append("%s")
                query.values.append(value)

            case ASTLogicalOperatorNode() as op:
                where.append(logical_operator_mapping[op])

            case ASTOperatorNode() as op:
                where.append(operator_mapping[op])

            case ASTComparisonNode(left, right, op):
                node_stack.append(iter((left, op, right)))

            case ASTGroupFlagNode.OPEN:
                if len(node_stack) > 1:
                    where.append("(")

            case ASTGroupFlagNode.CLOSE:
                if len(node_stack) > 1:
                    where.append(")")

            case None:
                node_stack.pop()

            case node:
                raise TypeError(f"Unexpected node type: {node}")

    query.where = " ".join(where)
    return query


def _process_ordering(sorting: list[ASTReferenceNode]) -> dict[str, ResultOrdering]:
    return {
        f"{ref.model.__model_name__}.{ref.field.name}": ref.ordering for ref in sorting
    }


def generate_joins(model: Type[OmmiModel], models: list[Type[OmmiModel]]):
    for join in models:
        yield create_join(model, join)


def create_join(model: Type[OmmiModel], join_model: Type[OmmiModel]) -> str:
    return f"JOIN {join_model.__ommi__.model_name} ON {create_join_comparison(model, join_model)}"


def create_join_comparison(model: Type[OmmiModel], join_model: Type[OmmiModel]) -> str:
    if model in join_model.__ommi__.references:
        columns = " AND ".join(
            f"{join_model.__ommi__.model_name}.{r.from_field.get('store_as')} = "
            f"{model.__ommi__.model_name}.{r.to_field.get('store_as')}"
            for r in join_model.__ommi__.references[model]
        )

    else:
        columns = " AND ".join(
            f"{join_model.__ommi__.model_name}.{r.to_field.get('store_as')} = "
            f"{model.__ommi__.model_name}.{r.from_field.get('store_as')}"
            for r in model.__ommi__.references[join_model]
        )

    return columns



import sqlite3

from ommi.drivers.database_results import async_result
from ommi.drivers.delete_actions import DeleteAction
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.ext.drivers.sqlite.utils import build_query, build_subquery
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode


class SQLiteDeleteAction(DeleteAction[SQLiteConnection, OmmiModel]):
    @async_result
    async def delete(self) -> bool:
        ast = when(*self._predicates)
        session = self._connection.cursor()
        self._delete(ast, session)
        return True

    def _delete(
        self,
        ast: ASTGroupNode,
        session: sqlite3.Cursor,
    ):
        query = build_query(ast)
        query_builder = ["DELETE FROM", query.model.__ommi__.model_name]
        if query.models:
            pks = ", ".join(
                f"{query.model.__ommi__.model_name}.{pk.get('store_as')}"
                for pk in query.model.get_primary_key_fields()
            )
            sub_query = build_subquery(query.model, query.models, query.where)
            query_builder.append(f"WHERE ({pks}) IN ({sub_query})")

        else:
            query_builder.append("WHERE")
            query_builder.append(query.where)

        session.execute(f"{' '.join(query_builder)};", query.values)



import sqlite3
from dataclasses import dataclass
from typing import Type, cast, TypeAlias

from ommi.drivers import DatabaseDriver, DriverConfig
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.drivers.drivers import enforce_connection_protocol, connection_context_manager
from ommi.ext.drivers.sqlite.add_action import SQLiteAddAction
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.ext.drivers.sqlite.find_action import SQLiteFindAction
from ommi.ext.drivers.sqlite.schema_action import SQLiteSchemaAction
from ommi.models.collections import ModelCollection
from ommi.models import OmmiModel
from ommi.query_ast import ASTGroupNode

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


@dataclass
class SQLiteConfig(DriverConfig):
    filename: str


@enforce_connection_protocol
class SQLiteDriver(
    DatabaseDriver[SQLiteConnection, OmmiModel],
    driver_name="sqlite",
    nice_name="SQLite",
):
    @async_result
    async def disconnect(self) -> bool:
        self._connection.close()
        self._connected = False
        return True

    @property
    def add(self) -> SQLiteAddAction:
        return SQLiteAddAction(self._connection)

    def find(self, *predicates: Predicate) -> SQLiteFindAction:
        return SQLiteFindAction(self._connection, predicates)

    def schema(
        self, model_collection: ModelCollection[Type[OmmiModel]] | None = None
    ) -> SQLiteSchemaAction:
        return SQLiteSchemaAction(self._connection, model_collection)

    @classmethod
    @connection_context_manager
    async def from_config(cls, config: SQLiteConfig) -> "SQLiteDriver":
        return cls(cast(SQLiteConnection, sqlite3.connect(config.filename)))



from typing import Protocol, runtime_checkable
from pymongo.client_session import ClientSession


@runtime_checkable
class MongoDBConnection(Protocol):
    def get_database(self, database_name: str):
        ...

    def close(self):
        ...

    async def start_session(self) -> ClientSession:
        ...



from functools import wraps
from typing import Generic, Type, TypeVar, NoReturn, Callable, Awaitable, ParamSpec

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")


class DatabaseResult(Generic[T]):
    Failure: "Type[DatabaseFailure[T]]"
    Success: "Type[DatabaseSuccess[T]]"

    @property
    def value(self) -> T | NoReturn:
        return

    @property
    def error(self) -> Exception | None:
        return

    def value_or(self, default: T) -> T:
        return

    def __bool__(self) -> bool:
        return False


class DatabaseSuccess(DatabaseResult[T]):
    __match_args__ = ("value",)

    def __init__(self, value: T):
        self._value = value

    @property
    def value(self):
        return self._value

    def value_or(self, default: T) -> T:
        return self._value

    def __bool__(self) -> bool:
        return True


class DatabaseFailure(DatabaseResult[T]):
    __match_args__ = ("error",)

    def __init__(self, error: Exception):
        self._error = error

    @property
    def value(self) -> NoReturn:
        raise RuntimeError("Cannot access value of a failed result") from self._error

    @property
    def error(self) -> Exception:
        return self._error

    def value_or(self, default: T) -> T:
        return default


DatabaseResult.Success = DatabaseSuccess
DatabaseResult.Failure = DatabaseFailure


class AsyncResultWrapper(Generic[T]):
    def __init__(self, awaitable):
        self._awaitable = awaitable

    def __await__(self):
        return self._await_and_wrap().__await__()

    async def raise_on_errors(self):
        match await self._await_and_wrap():
            case DatabaseResult.Failure(error):
                raise error

    @property
    async def value(self) -> T:
        return (await self._await_and_wrap()).value

    async def value_or(self, default: T) -> T:
        return (await self._await_and_wrap()).value_or(default)

    async def _await_and_wrap(self) -> DatabaseResult[T]:
        try:
            return DatabaseResult.Success(await self._awaitable)

        except Exception as error:
            return DatabaseResult.Failure(error)


def async_result(
    coroutine: Callable[P, Awaitable[R]]
) -> Callable[P, AsyncResultWrapper[R]]:
    @wraps(coroutine)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> AsyncResultWrapper[R]:
        return AsyncResultWrapper(coroutine(*args, **kwargs))

    return wrapper



from typing import Type, Iterable

from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.drivers.schema_actions import SchemaAction
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.models.collections import ModelCollection
from ommi.models import OmmiModel


class MongoDBSchemaAction(SchemaAction[MongoDBConnection, OmmiModel]):
    type_mapping = {
        int: "INTEGER",
        str: "TEXT",
        float: "REAL",
        bool: "INTEGER",
    }

    def __init__(
        self,
        connection: MongoDBConnection,
        model_collection: ModelCollection[Type[TModel]] | None,
        database,
    ):
        super().__init__(connection, model_collection)
        self._db = database

    @async_result
    async def create_models(self) -> Iterable[Type[OmmiModel]]:
        for model in self._model_collection.models:
            self._db.create_collection(model.__ommi__.model_name)

        return self._model_collection.models

    @async_result
    async def delete_models(self) -> None:
        for model in self._model_collection.models:
            await self._db[model.__ommi__.model_name].drop()



from abc import ABC, abstractmethod
from typing import Generic, Sequence, Type, TypeAlias

from ommi.drivers.driver_types import TConn, TModel

from ommi.drivers.database_results import AsyncResultWrapper
from ommi.query_ast import ASTGroupNode

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


class CountAction(Generic[TConn, TModel], ABC):
    def __init__(self, connection: TConn, predicates: Sequence[Predicate]):
        self._connection = connection
        self._predicates = predicates

    def __call__(self) -> AsyncResultWrapper[int]:
        return self.count()

    @abstractmethod
    def count(self) -> AsyncResultWrapper[int]:
        ...



from collections import defaultdict
from dataclasses import dataclass
from typing import Type, Any

from ommi import query_ast
from ommi.models.field_metadata import ReferenceTo, FieldMetadata
import ommi.models


@dataclass
class FieldReference:
    from_model: "Type[ommi.models.OmmiModel]"
    from_field: FieldMetadata
    to_model: "Type[ommi.models.OmmiModel]"
    to_field: FieldMetadata


class LazyReferenceBuilder:
    def __init__(
        self,
        fields: dict[str, FieldMetadata],
        model: "Type[ommi.models.OmmiModel]",
        namespace: dict[str, Any],
    ):
        self._built = False
        self._fields = fields
        self._model = model
        self._namespace = namespace
        self._references: dict[
            Type[ommi.models.OmmiModel], list[FieldReference]
        ] = defaultdict(list)

    def __contains__(self, model: "Type[ommi.models.OmmiModel]") -> bool:
        return model in self._references

    def __getitem__(self, model: "Type[ommi.models.OmmiModel]") -> list[FieldReference]:
        if not self._built:
            self._build_references()

        return self._references[model]

    def get(
        self,
        model: "Type[ommi.models.OmmiModel]",
        default: list[FieldReference] | None = None,
    ) -> list[FieldReference]:
        if not self._built:
            self._build_references()

        return self._references.get(model, default)

    def _build_references(self) -> None:
        for name, metadata in self._fields.items():
            if metadata.matches(ReferenceTo):
                match metadata.get("reference_to"):
                    case query_ast.ASTReferenceNode(to_field, to_model):
                        self._references[to_model].append(
                            FieldReference(
                                from_model=self._model,
                                from_field=metadata,
                                to_model=to_model,
                                to_field=to_field.metadata,
                            )
                        )

                    case str() as ref:
                        reference: query_ast.ASTReferenceNode = eval(
                            ref, vars(self._namespace)
                        )
                        self._references[reference.model].append(
                            FieldReference(
                                from_model=self._model,
                                from_field=metadata,
                                to_model=reference.model,
                                to_field=reference.field.metadata,
                            )
                        )

        self._built = True



from ommi.drivers.find_actions import FindAction
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.ext.drivers.sqlite.count_action import SQLiteCountAction
from ommi.ext.drivers.sqlite.delete_action import SQLiteDeleteAction
from ommi.ext.drivers.sqlite.fetch_action import SQLiteFetchAction
from ommi.ext.drivers.sqlite.set_fields_action import SQLiteSetFieldsAction
from ommi.models import OmmiModel


class SQLiteFindAction(FindAction[SQLiteConnection, OmmiModel]):
    _count_action = SQLiteCountAction
    _delete_action = SQLiteDeleteAction
    _fetch_action = SQLiteFetchAction
    _set_fields_action = SQLiteSetFieldsAction



from typing import Type

import psycopg

from ommi.drivers.database_results import async_result
from ommi.drivers.delete_actions import DeleteAction
from ommi.ext.drivers.postgresql.connection_protocol import PostgreSQLConnection
from ommi.ext.drivers.postgresql.utils import build_query, create_join_comparison
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode


class PostgreSQLDeleteAction(DeleteAction[PostgreSQLConnection, OmmiModel]):
    @async_result
    async def delete(self) -> bool:
        ast = when(*self._predicates)
        session = self._connection.cursor()
        await self._delete(ast, session)
        return self

    async def _delete(
        self,
        ast: ASTGroupNode,
        session: psycopg.AsyncCursor,
    ):
        query = build_query(ast)
        query_builder = ["DELETE FROM", query.model.__ommi__.model_name]
        where = [query.where]
        if query.models:
            query_builder.append("USING")
            query_builder.append(
                ", ".join(model.__ommi__.model_name for model in query.models)
            )
            using_join = [
                " AND ".join(
                    f"({create_join_comparison(query.model, model)})"
                    for model in query.models
                )
            ]

            where = []
            if query.where:
                where.append(f"({query.where}) AND")

            where.extend(using_join)

        query_builder.append("WHERE")
        query_builder.extend(where)
        await session.execute(f"{' '.join(query_builder)};".encode(), query.values)



from typing import Generic, Type, Sequence, TypeAlias
from abc import ABC, abstractmethod
from contextlib import suppress

from ommi.drivers.driver_types import TConn, TModel

from ommi.drivers.database_results import DatabaseResult, AsyncResultWrapper
import ommi.query_ast

Predicate: TypeAlias = "ommi.query_ast.ASTGroupNode | Type[TModel] | bool"


class FetchAction(Generic[TConn, TModel], ABC):
    def __init__(self, connection: TConn, predicates: Sequence[Predicate]):
        self._connection = connection
        self._predicates = predicates

    def __call__(self) -> AsyncResultWrapper[list[TModel]]:
        return self.fetch()

    @abstractmethod
    def fetch(self) -> AsyncResultWrapper[list[TModel]]:
        ...

    @abstractmethod
    async def one(self) -> TModel:
        ...

    async def all(self) -> list[TModel]:
        match await self.fetch():
            case DatabaseResult.Success(result):
                return result

            case DatabaseResult.Failure(error):
                raise error

    async def one_or(self, default: TModel) -> TModel:
        with suppress(Exception):
            return await self.one()

        return default



from typing import Generic, Iterable, Type
from abc import ABC, abstractmethod

from ommi.drivers.database_results import AsyncResultWrapper
from ommi.drivers.driver_types import TConn, TModel
from ommi.models.collections import ModelCollection


class SchemaAction(Generic[TConn, TModel], ABC):
    def __init__(
        self, connection: TConn, model_collection: ModelCollection[Type[TModel]] | None
    ):
        self._connection = connection
        self._model_collection = model_collection

    @abstractmethod
    def create_models(self) -> AsyncResultWrapper[Iterable[Type[TModel]]]:
        ...

    @abstractmethod
    def delete_models(self) -> AsyncResultWrapper[None]:
        ...



from typing import Type, TypeAlias
from dataclasses import dataclass

import motor.motor_asyncio

from ommi.drivers import DatabaseDriver, DriverConfig
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.drivers.drivers import enforce_connection_protocol, connection_context_manager
from ommi.ext.drivers.mongodb.add_action import MongoDBAddAction
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.ext.drivers.mongodb.find_action import MongoDBFindAction
from ommi.ext.drivers.mongodb.schema_action import MongoDBSchemaAction
from ommi.models.collections import ModelCollection
from ommi.models import OmmiModel
from ommi.query_ast import ASTGroupNode

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


@dataclass
class MongoDBConfig(DriverConfig):
    host: str
    port: int
    database_name: str
    timeout: int = 20000


@enforce_connection_protocol
class MongoDBDriver(
    DatabaseDriver[MongoDBConnection, OmmiModel],
    driver_name="mongodb",
    nice_name="MongoDB",
):
    def __init__(self, connection: MongoDBConnection, database):
        super().__init__(connection)
        self._db = database

    @async_result
    async def disconnect(self) -> bool:
        self._connection.close()
        self._connected = False
        return True

    @property
    def add(self) -> MongoDBAddAction:
        return MongoDBAddAction(self._connection, self._db)

    def find(self, *predicates: Predicate) -> MongoDBFindAction:
        return MongoDBFindAction(self._connection, predicates, self._db)

    def schema(
        self, model_collection: ModelCollection[Type[OmmiModel]] | None = None
    ) -> MongoDBSchemaAction:
        return MongoDBSchemaAction(self._connection, model_collection, self._db)

    @classmethod
    @connection_context_manager
    async def from_config(cls, config: MongoDBConfig) -> "MongoDBDriver":
        connection = motor.motor_asyncio.AsyncIOMotorClient(
            config.host, config.port, timeoutMS=config.timeout
        )
        db = connection.get_database(config.database_name)
        return cls(connection, db)



import sqlite3
from typing import Any

from ommi.drivers.database_results import async_result
from ommi.drivers.set_fields_actions import SetFieldsAction
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.ext.drivers.sqlite.utils import build_query, build_subquery
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode


class SQLiteSetFieldsAction(SetFieldsAction[SQLiteConnection, OmmiModel]):
    @async_result
    async def set_fields(self, **kwargs: Any) -> bool:
        ast = when(*self._predicates)
        session = self._connection.cursor()
        self._update(ast, kwargs, session)
        return True

    def _update(
        self,
        ast: ASTGroupNode,
        set_fields: dict[str, Any],
        session: sqlite3.Cursor,
    ):
        query = build_query(ast)
        fields = query.model.__ommi__.fields
        query_builder = [
            f"UPDATE {query.model.__ommi__.model_name}",
            f"SET",
            ", ".join(
                f"{fields[name].get('store_as')} = ?" for name in set_fields.keys()
            ),
        ]
        if query.models:
            sub_query = build_subquery(query.model, query.models, query.where)
            pks = ", ".join(
                f"{query.model.__ommi__.model_name}.{pk.get('store_as')}"
                for pk in query.model.get_primary_key_fields()
            )
            query_builder.append(f"WHERE ({pks}) IN ({sub_query})")

        else:
            query_builder.append("WHERE")
            query_builder.append(query.where)

        session.execute(
            f"{' '.join(query_builder)};",
            (*set_fields.values(), *query.values),
        )



from contextlib import suppress

from typing import Iterable

from ommi.drivers.add_actions import AddAction
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.ext.drivers.mongodb.utils import model_to_dict
from ommi.models import OmmiModel


class MongoDBAddAction(AddAction[MongoDBConnection, OmmiModel]):
    def __init__(self, connection: MongoDBConnection, database):
        super().__init__(connection)
        self._db = database

    @async_result
    async def items(self, *items: TModel) -> Iterable[TModel]:
        for item in items:
            await self._insert(item)

        return items

    async def _insert(self, item: OmmiModel):
        data = model_to_dict(item)
        result = await self._db[item.__ommi__.model_name].insert_one(data)
        item.__ommi_mongodb_id__ = result.inserted_id
        await self._set_auto_increment_pk(item)

    async def _set_auto_increment_pk(self, item: OmmiModel):
        pks = item.get_primary_key_fields()
        if len(pks) != 1:
            return

        pk = pks[0]
        if getattr(item, pk.get("field_name")) is not None:
            return

        if (
            not issubclass(pk.get("field_type"), int)
            or getattr(item, pk.get("field_name")) is not None
        ):
            return

        name = pk.get("store_as")
        with suppress(StopAsyncIteration):
            await self._db[item.__ommi__.model_name].aggregate(
                [
                    {
                        "$lookup": {
                            "from": item.__ommi__.model_name,
                            "pipeline": [
                                {
                                    "$match": {
                                        name: {
                                            "$exists": True,
                                        },
                                    },
                                },
                                {
                                    "$sort": {
                                        name: 1,
                                    },
                                },
                                {
                                    "$addFields": {
                                        "next_id": {
                                            "$add": [f"${name}", 1],
                                        },
                                    },
                                },
                            ],
                            "as": "__ommi_autoincrement",
                        },
                    },
                    {
                        "$unwind": {
                            "path": "$__ommi_autoincrement",
                            "preserveNullAndEmptyArrays": True,
                        },
                    },
                    {
                        "$match": {
                            "_id": item.__ommi_mongodb_id__,
                        },
                    },
                    {
                        "$addFields": {
                            name: {
                                "$ifNull": [
                                    "$__ommi_autoincrement.next_id",
                                    0,
                                ],
                            },
                        },
                    },
                    {
                        "$unset": ["__ommi_autoincrement"],
                    },
                    {
                        "$match": {
                            name: {
                                "$exists": True,
                            },
                        },
                    },
                    {
                        "$merge": item.__ommi__.model_name,
                    },
                ]
            ).next()

        result = await self._db[item.__ommi__.model_name].find_one(
            {"_id": item.__ommi_mongodb_id__}, {"_id": 0, name: 1}
        )
        setattr(item, pk.get("field_name"), result[name])



import sqlite3
from typing import TypeAlias, Type

from ommi.drivers.count_actions import CountAction
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode, ResultOrdering
from ommi.ext.drivers.sqlite.utils import build_query, SelectQuery, generate_joins

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


class SQLiteCountAction(CountAction[SQLiteConnection, OmmiModel]):
    @async_result
    async def count(self) -> int:
        ast = when(*self._predicates)
        session = self._connection.cursor()
        return self._count(ast, session)

    def _count(self, predicates: ASTGroupNode, session: sqlite3.Cursor):
        query = build_query(predicates)
        query_str = self._build_count_query(query)
        session.execute(query_str, query.values)
        result = session.fetchone()
        return result[0]

    def _build_count_query(self, query: SelectQuery):
        query_builder = [f"SELECT Count(*) FROM {query.model.__ommi__.model_name}"]
        if query.models:
            query_builder.extend(generate_joins(query.model, query.models))

        if query.where:
            query_builder.append(f"WHERE {query.where}")

        if query.limit > 0:
            query_builder.append(f"LIMIT {query.limit}")

            if query.offset > 0:
                query_builder.append(f"OFFSET {query.offset}")

        if query.order_by:
            ordering = ", ".join(
                f"{column} {'ASC' if ordering is ResultOrdering.ASCENDING else 'DESC'}"
                for column, ordering in query.order_by.items()
            )
            query_builder.append("ORDER BY")
            query_builder.append(ordering)

        return " ".join(query_builder) + ";"



from typing import Type, Iterable, Any

import psycopg

from ommi.drivers.database_results import async_result
from ommi.drivers.schema_actions import SchemaAction
from ommi.ext.drivers.postgresql.connection_protocol import PostgreSQLConnection
from ommi.models.field_metadata import FieldMetadata
from ommi.models import OmmiModel
from ommi.models.models import get_collection
from tramp.optionals import Optional


class PostgreSQLSchemaAction(SchemaAction[PostgreSQLConnection, OmmiModel]):
    type_mapping = {
        int: "INTEGER",
        str: "TEXT",
        float: "REAL",
        bool: "BOOLEAN",
    }

    @async_result
    async def create_models(self) -> Iterable[Type[OmmiModel]]:
        session = self._connection.cursor()
        models = get_collection(
            Optional.Some(self._model_collection)
            if self._model_collection
            else Optional.Nothing
        ).models

        try:
            for model in models:
                await self._create_table(model, session)

        except:
            await self._connection.rollback()
            raise

        else:
            return models

        finally:
            await session.close()

    @async_result
    async def delete_models(self) -> None:
        session = self._connection.cursor()
        models = get_collection(
            Optional.Some(self._model_collection)
            if self._model_collection
            else Optional.Nothing
        ).models

        try:
            for model in models:
                await session.execute(
                    f"DROP TABLE IF EXISTS {model.__ommi__.model_name};"
                )

        except:
            await self._connection.rollback()
            raise

        else:
            return None

        finally:
            await session.close()

    async def _create_table(self, model: Type[OmmiModel], session: psycopg.AsyncCursor):
        pks = model.get_primary_key_fields()
        columns = ", ".join(
            self._build_column(field, len(pks) == 1 and field in pks)
            for field in model.__ommi__.fields.values()
        )
        await session.execute(
            f"CREATE TABLE IF NOT EXISTS {model.__ommi__.model_name}"
            f"("
            f"{columns}, "
            f"PRIMARY KEY ({', '.join(pk.get('store_as') for pk in pks)})"
            f");"
        )

    def _build_column(self, field: FieldMetadata, pk: bool) -> str:
        column = [
            field.get("store_as"),
            self._get_postgresql_type(field.get("field_type"), pk),
        ]
        return " ".join(column)

    def _get_postgresql_type(self, type_: Type[Any], pk: bool) -> str:
        if pk and isinstance(type_, type) and issubclass(type_, int):
            return "SERIAL"

        return self.type_mapping.get(type_, "TEXT")



from typing import Type, Any, TypeVar, Sequence

from ommi.drivers.database_results import async_result
from ommi.drivers.fetch_actions import FetchAction
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.ext.drivers.mongodb.utils import build_pipeline, process_ast
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode


T = TypeVar("T")
Predicate = ASTGroupNode | Type[OmmiModel] | bool


class Iteratable:
    pass


class MongoDBFetchAction(FetchAction[MongoDBConnection, OmmiModel]):
    def __init__(
        self, connection: MongoDBConnection, predicates: Sequence[Predicate], database
    ):
        super().__init__(connection, predicates)
        self._db = database

    @async_result
    async def fetch(self) -> list[OmmiModel]:
        query = process_ast(when(*self._predicates))
        pipeline, model = build_pipeline(query)
        results = self._db[model.__ommi__.model_name].aggregate(pipeline)
        return [self._create_model(result, model) async for result in results]

    async def one(self) -> OmmiModel:
        return (await self.all())[0]

    def _create_model(self, data: dict[str, Any], model: Type[OmmiModel]) -> OmmiModel:
        field_mapping = {
            field.get("store_as"): field.get("field_name")
            for field in model.__ommi__.fields.values()
        }
        instance = model(
            **{
                field_mapping[key]: value
                for key, value in data.items()
                if key in field_mapping
            }
        )
        instance.__ommi_mongodb_id__ = data.get("_id")
        return instance



from typing import Generic, Sequence, Type, TypeAlias
from abc import ABC, abstractmethod

from ommi.drivers.driver_types import TConn, TModel

from ommi.drivers.database_results import AsyncResultWrapper
import ommi.query_ast

Predicate: TypeAlias = "ommi.query_ast.ASTGroupNode | Type[TModel] | bool"


class DeleteAction(Generic[TConn, TModel], ABC):
    def __init__(self, connection: TConn, predicates: Sequence[Predicate]):
        self._connection = connection
        self._predicates = predicates

    def __call__(self) -> AsyncResultWrapper[bool]:
        return self.delete()

    @abstractmethod
    def delete(self) -> AsyncResultWrapper[bool]:
        ...



from abc import ABC, abstractmethod
from typing import Generic, Iterable

from ommi.drivers.driver_types import TConn, TModel
from ommi.drivers.database_results import AsyncResultWrapper


class AddAction(Generic[TConn, TModel], ABC):
    def __init__(self, connection: TConn):
        self._connection = connection

    def __call__(self, *items: TModel) -> AsyncResultWrapper[Iterable[TModel]]:
        return self.items(*items)

    @abstractmethod
    def items(self, *items: TModel) -> AsyncResultWrapper[Iterable[TModel]]:
        ...



from ommi import query_ast


class QueryableFieldDescriptor:
    def __init__(self, field, metadata):
        self.field = field
        self.metadata = metadata

    def __get__(self, instance, owner):
        if instance is None:
            return query_ast.ASTReferenceNode(self, owner)

        return self.field



from typing import Iterable, Type, Sequence

import psycopg

from ommi.drivers.add_actions import AddAction
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.ext.drivers.postgresql.connection_protocol import PostgreSQLConnection
from ommi.models import OmmiModel


class PostgreSQLAddAction(AddAction[PostgreSQLConnection, OmmiModel]):
    @async_result
    async def items(self, *items: TModel) -> Iterable[TModel]:
        session = self._connection.cursor()
        try:
            for model, group in self._group_by_model_type(items).items():
                await self._insert(group, session, model)

        except Exception:
            await self._connection.rollback()
            raise

        else:
            return items

        finally:
            await session.close()

    def _group_by_model_type(self, items: Sequence[OmmiModel]):
        model_groups = {}
        for item in items:
            model_groups.setdefault(type(item), []).append(item)

        return model_groups

    async def _insert(
        self,
        items: Sequence[OmmiModel],
        session: psycopg.AsyncCursor,
        model: Type[OmmiModel],
    ):
        query = [f"INSERT INTO {model.__ommi__.model_name}"]

        fields = list(model.__ommi__.fields.values())
        pks = {
            pk: getattr(items[0], pk.get("field_name")) is not None
            for pk in model.get_primary_key_fields()
        }
        columns = [field.get("store_as") for field in fields if pks.get(field, True)]
        query.append(f"({','.join(columns)})")

        values = []
        inserts = []
        for item in items:
            qs = ",".join(["%s"] * len(columns))
            inserts.append(f"({qs})")
            values.extend(
                getattr(item, field.get("field_name"))
                for field in fields
                if pks.get(field, True)
            )

        query.append(f"VALUES {','.join(inserts)}")
        query.append(f"RETURNING {', '.join(pk.get('store_as') for pk in pks)};")

        result = await session.execute(" ".join(query).encode(), values)

        # Update the primary key field of the models that were inserted if the primary key is an auto-incrementing field
        item_stack = iter(items)
        async for record in result:
            item = next(item_stack)
            for pk, value in zip(pks, record):
                setattr(item, pk.get("field_name"), value)



import sys
from dataclasses import dataclass
from inspect import get_annotations
from typing import (
    Callable,
    overload,
    Type,
    TypeVar,
    Any,
    Generator,
    Annotated,
)
from tramp.annotations import Annotation
from tramp.optionals import Optional

import ommi.query_ast as query_ast
import ommi
from ommi.drivers.database_results import async_result, AsyncResultWrapper
from ommi.models.field_metadata import (
    FieldMetadata,
    FieldType,
    StoreAs,
    create_metadata_type,
    AggregateMetadata,
    Key,
)
from ommi.models.metadata import OmmiMetadata
from ommi.contextual_method import contextual_method
import ommi.models.collections

import ommi.drivers.delete_actions as delete_actions
import ommi.drivers.fetch_actions as fetch_actions
import ommi.models.query_fields
from ommi.models.queryable_descriptors import QueryableFieldDescriptor
from ommi.models.references import LazyReferenceBuilder
from ommi.utils.get_first import first

try:
    from typing import Self
except ImportError:
    Self = Any

T = TypeVar("T", bound=Type)

DRIVER_DUNDER_NAME = "__ommi_driver__"
MODEL_NAME_DUNDER_NAME = "__ommi_model_name__"
MODEL_NAME_CLASS_PARAM = "name"
METADATA_DUNDER_NAME = "__ommi__"


def _get_value(
    class_params: dict[str, Any],
    param_name: str,
    cls: Type[Any],
    dunder_name: str,
    default: Any,
) -> Any:
    return class_params.pop(param_name, getattr(cls, dunder_name, default))


@dataclass
class QueryFieldMetadata:
    name: str
    type: "Type[ommi.models.query_fields.LazyQueryField]"
    args: tuple[Any, ...]


class OmmiModel:
    __ommi__: OmmiMetadata

    @contextual_method
    def get_driver(
        self, driver: "drivers.DatabaseDrivers | None" = None
    ) -> "drivers.DatabaseDriver | None":
        return driver or type(self).get_driver()

    @get_driver.classmethod
    def get_driver(
        cls, driver: "drivers.DatabaseDrivers | None" = None
    ) -> "drivers.DatabaseDriver | None":
        return driver or ommi.active_driver.get(None)

    def delete(
        self, driver: "drivers.DatabaseDriver | None" = None
    ) -> "delete_actions.DeleteAction":
        return (
            self.get_driver(driver)
            .find(
                query_ast.when(
                    *(
                        getattr(type(self), pk.get("field_name"))
                        == getattr(self, pk.get("field_name"))
                        for pk in self.get_primary_key_fields()
                    )
                )
            )
            .delete()
        )

    @classmethod
    def count(
        cls,
        *predicates: "ASTGroupNode | DatabaseModel | bool",
        columns: Any | None = None,
        driver: "drivers.DatabaseDriver | None" = None,
    ) -> AsyncResultWrapper[int]:
        driver = cls.get_driver(driver)
        if not predicates and not columns:
            predicates = (cls,)

        return driver.find(*predicates, *cls._build_column_predicates(columns)).count()

    @classmethod
    def fetch(
        cls,
        *predicates: "ASTGroupNode | DatabaseModel | bool",
        driver: "drivers.DatabaseDriver | None" = None,
        **columns: Any,
    ) -> "fetch_actions.FetchAction[OmmiModel]":
        driver = cls.get_driver(driver)
        return driver.find(*predicates, *cls._build_column_predicates(columns)).fetch

    @async_result
    async def reload(self, driver: "drivers.DatabaseDriver | None" = None) -> Self:
        result = await (
            self.get_driver(driver)
            .find(
                query_ast.when(
                    *(
                        getattr(type(self), pk.get("field_name"))
                        == getattr(self, pk.get("field_name"))
                        for pk in self.get_primary_key_fields()
                    )
                )
            )
            .fetch.one()
        )
        for name in self.__ommi__.fields.keys():
            setattr(self, name, getattr(result, name))

        return self

    @async_result
    async def save(self, driver: "drivers.DatabaseDriver | None" = None) -> bool:
        pks = self.get_primary_key_fields()
        driver = self.get_driver(driver)
        await driver.find(
            query_ast.when(
                *(
                    getattr(type(self), pk.get("field_name"))
                    == getattr(self, pk.get("field_name"))
                    for pk in pks
                )
            )
        ).set(
            **{
                field.get("field_name"): getattr(self, field.get("field_name"))
                for field in self.__ommi__.fields.values()
                if field not in pks
                and getattr(self, field.get("field_name")) is not None
            }
        )
        return True

    @classmethod
    def get_primary_key_fields(cls) -> tuple[FieldMetadata, ...]:
        fields = cls.__ommi__.fields
        if not fields:
            raise Exception(f"No fields defined on {cls}")

        def find_fields_where(predicate):
            return tuple(f for f in fields.values() if predicate(f))

        def find_field_where(predicate):
            return first(find_fields_where(predicate))

        if matches := find_fields_where(lambda f: f.matches(Key)):
            return matches

        if field := find_field_where(lambda f: f.get("store_as") in {"id", "_id"}):
            return (field,)

        if field := find_field_where(lambda f: issubclass(f.get("field_type"), int)):
            return (field,)

        return (first(fields.values()),)

    @classmethod
    def _build_column_predicates(
        cls, columns: dict[str, Any]
    ) -> "Generator[query_ast.ASTComparisonNode | bool, None, None]":
        if not columns:
            return

        for name, value in columns.items():
            if name not in cls.__ommi__.fields:
                raise ValueError(f"Invalid column {name!r} for model {cls.__name__}")

            yield getattr(cls, name) == value


@overload
def ommi_model(
    cls: None = None,
    *,
    collection: "ommi.model_collections.ModelCollection | None" = None,
) -> Callable[[T], T | Type[OmmiModel]]:
    ...


@overload
def ommi_model(cls: T) -> T | Type[OmmiModel]:
    ...


def ommi_model(
    cls: T | None = None, /, **kwargs
) -> T | Type[OmmiModel] | Callable[[T], T | Type[OmmiModel]]:
    def wrap_model(c: T) -> T | Type[OmmiModel]:
        model = _create_model(c, **kwargs)
        _register_model(
            model,
            (
                Optional.Some(kwargs["collection"])
                if "collection" in kwargs
                else Optional.Nothing
            ),
        )
        return model

    return wrap_model if cls is None else wrap_model(cls)


def _create_model(c: T, **kwargs) -> T | Type[OmmiModel]:
    metadata_factory = (
        c.__ommi__.clone if hasattr(c, METADATA_DUNDER_NAME) else OmmiMetadata
    )

    annotations = {
        name: Annotation(hint, Optional.Some(vars(sys.modules[c.__module__])))
        for name, hint in get_annotations(c).items()
    }
    fields = _get_fields(annotations)
    query_fields = _get_query_fields(annotations)

    def init(self, *args, **kwargs):
        unset_query_fields = {name: None for name in query_fields if name not in kwargs}
        super(model_type, self).__init__(*args, **kwargs | unset_query_fields)

        for name, annotation in query_fields.items():
            if name in unset_query_fields:
                setattr(self, name, annotation.origin.create(self, annotation.args))

    model_type = type.__new__(
        type(c),
        f"OmmiModel_{c.__name__}",
        (c, OmmiModel),
        {
            name: QueryableFieldDescriptor(getattr(c, name, None), fields[name])
            for name in fields
        }
        | {
            "__init__": init,
            METADATA_DUNDER_NAME: metadata_factory(
                model_name=_get_value(
                    kwargs,
                    MODEL_NAME_CLASS_PARAM,
                    c,
                    MODEL_NAME_DUNDER_NAME,
                    c.__name__,
                ),
                fields=fields,
                references=LazyReferenceBuilder(fields, c, sys.modules[c.__module__]),
            ),
        },
    )
    getattr(model_type, METADATA_DUNDER_NAME).references._model = model_type
    return model_type


def _register_model(
    model: Type[OmmiModel],
    collection: "Optional[ommi.models.collections.ModelCollection]",
):
    get_collection(collection, model).add(model)


def get_collection(
    collection: "Optional[ommi.models.collections.ModelCollection]",
    model: Type[OmmiModel] | None = None,
) -> "ommi.models.collections.ModelCollection":
    return collection.value_or(
        getattr(model, METADATA_DUNDER_NAME).collection
        if model
        else ommi.models.collections.get_global_collection()
    )


def _get_fields(fields: dict[str, Annotation]) -> dict[str, FieldMetadata]:
    ommi_fields = {}
    for name, annotation in fields.items():
        metadata = AggregateMetadata()

        if annotation.origin == Annotated:
            for arg in annotation.args:
                match arg:
                    case FieldMetadata():
                        metadata |= arg

        if not issubclass(annotation.type, ommi.models.query_fields.LazyQueryField):
            ommi_fields[name] = metadata | FieldType(annotation.type)
            if not ommi_fields[name].matches(StoreAs):
                ommi_fields[name] |= StoreAs(name)

            ommi_fields[name] |= create_metadata_type(
                "FieldMetadata", field_name=name, field_type=annotation.type
            )()

    return ommi_fields


def _get_query_fields(fields: dict[str, Annotation]) -> dict[str, Annotation]:
    return {
        name: annotation
        for name, annotation in fields.items()
        if annotation.is_generic()
        and issubclass(annotation.type, ommi.models.query_fields.LazyQueryField)
    }



import sqlite3
from datetime import datetime, date
from typing import Type, Any, Generator, TypeVar, Callable, get_origin

from ommi.drivers.database_results import async_result
from ommi.drivers.fetch_actions import FetchAction
from ommi.ext.drivers.sqlite.connection_protocol import SQLiteConnection
from ommi.ext.drivers.sqlite.utils import build_query, SelectQuery, generate_joins
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode, ResultOrdering


T = TypeVar("T")


class Iteratable:
    pass


class SQLiteFetchAction(FetchAction[SQLiteConnection, OmmiModel]):
    type_validators = {
        datetime: lambda value: value,  # No-op to avoid type conflict with date
        date: lambda value: value.split()[0],
    }

    @async_result
    async def fetch(self) -> list[OmmiModel]:
        ast = when(*self._predicates)
        session = self._connection.cursor()
        result = self._select(ast, session)
        return result

    async def one(self) -> OmmiModel:
        return (await self.all())[0]

    def _select(self, predicates: ASTGroupNode, session: sqlite3.Cursor):
        query = build_query(predicates)
        query_str = self._build_select_query(query)
        session.execute(query_str, query.values)
        result = session.fetchall()
        return [
            query.model(**dict(self._validate_row_values(query.model, row)))
            for row in result
        ]

    def _build_select_query(self, query: SelectQuery):
        query_builder = [f"SELECT * FROM {query.model.__ommi__.model_name}"]
        if query.models:
            query_builder.extend(generate_joins(query.model, query.models))

        if query.where:
            query_builder.append(f"WHERE {query.where}")

        if query.limit > 0:
            query_builder.append(f"LIMIT {query.limit}")

            if query.offset > 0:
                query_builder.append(f"OFFSET {query.offset}")

        if query.order_by:
            ordering = ", ".join(
                f"{column} {'ASC' if ordering is ResultOrdering.ASCENDING else 'DESC'}"
                for column, ordering in query.order_by.items()
            )
            query_builder.append("ORDER BY")
            query_builder.append(ordering)

        return " ".join(query_builder) + ";"

    def _validate_row_values(
        self, model: Type[OmmiModel], row: tuple[Any]
    ) -> Generator[tuple[str, Any], None, None]:
        for field, value in zip(model.__ommi__.fields.values(), row):
            name = field.get("field_name")
            if validator := self._find_type_validator(field.get("field_type", value)):
                yield name, validator(value)
            else:
                yield name, value

    def _find_type_validator(self, type_hint: Type[T]) -> Callable[[Any], T] | None:
        hint = get_origin(type_hint) or type_hint
        for validator_type, validator in self.type_validators.items():
            if issubclass(hint, validator_type):
                return validator

        return None



from typing import TypeAlias, Type, Sequence

from ommi.drivers.count_actions import CountAction
from ommi.drivers.database_results import async_result
from ommi.drivers.driver_types import TModel
from ommi.ext.drivers.mongodb.connection_protocol import MongoDBConnection
from ommi.models import OmmiModel
from ommi.query_ast import when, ASTGroupNode
from ommi.ext.drivers.mongodb.utils import build_pipeline, process_ast

Predicate: TypeAlias = ASTGroupNode | Type[TModel] | bool


class MongoDBCountAction(CountAction[MongoDBConnection, OmmiModel]):
    def __init__(
        self, connection: MongoDBConnection, predicates: Sequence[Predicate], database
    ):
        super().__init__(connection, predicates)
        self._db = database

    @async_result
    async def count(self) -> int:
        query = process_ast(when(*self._predicates))
        pipeline, model = build_pipeline(query)
        pipeline.append({"$count": "count"})
        result = (
            await self._db[model.__ommi__.model_name].aggregate(pipeline).to_list(1)
        )
        return result[0].get("count", 0)
